# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the fetchai module."""
import base64
import json
import logging
import shutil
import tempfile
import time
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock, call, patch
from uuid import uuid4

import pytest
from aea_ledger_fetchai import FetchAIApi, FetchAICrypto, FetchAIFaucetApi
from cosmpy.protos.cosmos.bank.v1beta1.tx_pb2 import MsgSend
from cosmpy.protos.cosmos.base.v1beta1.coin_pb2 import Coin
from google.protobuf.any_pb2 import Any as ProtoAny

from aea.crypto.helpers import KeyIsIncorrect

from tests.conftest import FETCHAI_TESTNET_CONFIG, MAX_FLAKY_RERUNS, ROOT_DIR


@pytest.fixture
def fetchai_private_key_file():
    """Pytest fixture to create a temporary FetchAI private key file."""
    crypto = FetchAICrypto()
    temp_dir = Path(tempfile.mkdtemp())
    try:
        temp_file = temp_dir / "private.key"
        temp_file.write_text(crypto.private_key)
        yield str(temp_file)
    finally:
        shutil.rmtree(temp_dir)


class MockRequestsResponse:
    """Mock of request response."""

    def __init__(self, data, status_code=None):
        """Initialize mock of request response."""
        self._data = data
        self._status_code = status_code or 200

    @property
    def status_code(self):
        """Get status code."""
        return 200

    def json(self):
        """Get json."""
        return self._data


def test_creation(fetchai_private_key_file):
    """Test the creation of the crypto_objects."""
    assert FetchAICrypto(), "Did not manage to initialise the crypto module"
    assert FetchAICrypto(
        fetchai_private_key_file
    ), "Did not manage to load the cosmos private key"


def test_key_file_encryption_decryption(fetchai_private_key_file):
    """Test fetchai private key encrypted and decrypted correctly."""
    fetchai = FetchAICrypto(fetchai_private_key_file)
    pk_data = Path(fetchai_private_key_file).read_text()
    password = uuid4().hex
    encrypted_data = fetchai.encrypt(password)
    decrypted_data = fetchai.decrypt(encrypted_data, password)
    assert encrypted_data != pk_data
    assert pk_data == decrypted_data

    with pytest.raises(ValueError, match="Decrypt error! Bad password?"):
        fetchai.decrypt(encrypted_data, "BaD_PassWord")

    with pytest.raises(ValueError, match="Bad encrypted key format!"):
        fetchai.decrypt("some_data" * 16, "BaD_PassWord")


def test_initialization():
    """Test the initialisation of the variables."""
    account = FetchAICrypto()
    assert account.entity is not None, "The property must return the account."
    assert (
        account.address is not None
    ), "After creation the display address must not be None"
    assert account.address.startswith("fetch")
    assert (
        account.public_key is not None
    ), "After creation the public key must no be None"


def test_sign_and_recover_message(fetchai_private_key_file):
    """Test the signing and the recovery of a message."""
    account = FetchAICrypto(fetchai_private_key_file)
    sign_bytes = account.sign_message(message=b"hello")
    assert len(sign_bytes) > 0, "The len(signature) must not be 0"
    recovered_addresses = FetchAIApi.recover_message(
        message=b"hello", signature=sign_bytes
    )
    assert (
        account.address in recovered_addresses
    ), "Failed to recover the correct address."


def test_get_hash():
    """Test the get hash functionality."""
    expected_hash = "2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824"
    hash_ = FetchAIApi.get_hash(message=b"hello")
    assert expected_hash == hash_


def test_dump_positive(fetchai_private_key_file):
    """Test dump."""
    account = FetchAICrypto(fetchai_private_key_file)
    account.dump(MagicMock())


def test_api_creation():
    """Test api instantiation."""
    assert FetchAIApi(**FETCHAI_TESTNET_CONFIG), "Failed to initialise the api"


def test_api_none():
    """Test the "api" of the cryptoApi is none."""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    assert fetchai_api.api is None, "The api property is not None."


def test_generate_nonce():
    """Test generate nonce."""
    nonce = FetchAIApi.generate_tx_nonce(
        seller="some_seller_addr", client="some_buyer_addr"
    )
    assert len(nonce) > 0 and int(
        nonce, 16
    ), "The len(nonce) must not be 0 and must be hex"


def test_get_address_from_public_key():
    """Test the address from public key."""
    fet_crypto = FetchAICrypto()
    address = FetchAIApi.get_address_from_public_key(fet_crypto.public_key)
    assert address == fet_crypto.address, "The address must be the same."


def test_validate_address():
    """Test the is_valid_address functionality."""
    account = FetchAICrypto()
    assert FetchAIApi.is_valid_address(account.address)
    assert not FetchAIApi.is_valid_address(account.address + "wrong")


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_construct_sign_and_submit_transfer_transaction():
    """Test the construction, signing and submitting of a transfer transaction."""
    account = FetchAICrypto()
    balance = get_wealth(account.address)
    assert balance > 0, "Failed to fund account."
    fc2 = FetchAICrypto()
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    amount = 10000
    assert amount < balance, "Not enough funds."
    transfer_transaction = fetchai_api.get_transfer_transaction(
        sender_address=account.address,
        destination_address=fc2.address,
        amount=amount,
        tx_fee=1550000000000000,
        tx_nonce="something",
    )
    assert (
        isinstance(transfer_transaction, dict) and len(transfer_transaction) == 2
    ), "Incorrect transfer_transaction constructed."

    signed_transaction = account.sign_transaction(transfer_transaction)
    assert (
        isinstance(signed_transaction, dict)
        and len(signed_transaction["tx"]) == 3
        and isinstance(signed_transaction["tx"]["signatures"], list)
    ), "Incorrect signed_transaction constructed."

    transaction_digest = fetchai_api.send_signed_transaction(signed_transaction)
    assert transaction_digest is not None, "Failed to submit transfer transaction!"

    not_settled = True
    elapsed_time = 0
    while not_settled and elapsed_time < 20:
        elapsed_time += 1
        time.sleep(2)
        transaction_receipt = fetchai_api.get_transaction_receipt(transaction_digest)
        if transaction_receipt is None:
            continue
        is_settled = fetchai_api.is_transaction_settled(transaction_receipt)
        not_settled = not is_settled
    assert transaction_receipt is not None, "Failed to retrieve transaction receipt."
    assert is_settled, "Failed to verify tx!"

    tx = fetchai_api.get_transaction(transaction_digest)
    is_valid = fetchai_api.is_transaction_valid(
        tx, fc2.address, account.address, "", amount
    )
    assert is_valid, "Failed to settle tx correctly!"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_balance():
    """Test the balance is zero for a new account."""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    fc = FetchAICrypto()
    balance = fetchai_api.get_balance(fc.address)
    assert balance == 0, "New account has a positive balance."
    balance = get_wealth(fc.address)
    assert balance > 0, "Existing account has no balance."


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_state():
    """Test that get_state() with 'blocks' function returns something containing the block height."""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    callable_name = "blocks"
    args = ("latest",)
    block = fetchai_api.get_state(callable_name, *args)
    assert block is not None, "No response to 'blocks/latest' query."
    assert (
        block["block"]["header"]["height"] is not None
    ), "Block height not found in response."


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def get_wealth(address: str):
    """Get wealth for test."""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    FetchAIFaucetApi().get_wealth(address)
    balance = 0
    timeout = 0
    while timeout < 40 and balance == 0:
        time.sleep(1)
        timeout += 1
        _balance = fetchai_api.get_balance(address)
        balance = _balance if _balance is not None else 0
    return balance


# @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_wealth_positive(caplog):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.fetchai._default_logger"):
        fetchai_faucet_api = FetchAIFaucetApi()
        fc = FetchAICrypto()
        fetchai_faucet_api.get_wealth(fc.address)


@pytest.mark.ledger
@mock.patch("aea_ledger_cosmos.cosmos.requests.get")
@mock.patch("aea_ledger_cosmos.cosmos.requests.post")
def test_successful_faucet_operation(mock_post, mock_get):
    """Test successful faucet operation."""
    address = "a normal cosmos address would be here"
    mock_post.return_value = MockRequestsResponse({"uuid": "a-uuid-v4-would-be-here"})

    mock_get.return_value = MockRequestsResponse(
        {
            "status": "ok",
            "claim": {
                "createdAt": "2021-08-13T15:18:50.420Z",
                "updatedAt": "2021-08-13T15:18:58.249Z",
                "status": FetchAIFaucetApi.FAUCET_STATUS_COMPLETED,
                "txStatus": {
                    "hash": "0x transaction hash would be here",
                    "height": 123456,
                },
            },
        }
    )

    faucet = FetchAIFaucetApi()
    faucet.get_wealth(address)

    mock_post.assert_has_calls(
        [
            call(
                url=f"{FetchAIFaucetApi.testnet_faucet_url}/api/v3/claims",
                json={"address": address},
            )
        ]
    )
    mock_get.assert_has_calls(
        [
            call(
                f"{FetchAIFaucetApi.testnet_faucet_url}/api/v3/claims/a-uuid-v4-would-be-here"
            )
        ]
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.ledger
@mock.patch("aea_ledger_cosmos.cosmos.requests.get")
@mock.patch("aea_ledger_cosmos.cosmos.requests.post")
def test_successful_realistic_faucet_operation(mock_post, mock_get):
    """Test successful realistic faucet operation."""
    address = "a normal cosmos address would be here"
    mock_post.return_value = MockRequestsResponse({"uuid": "a-uuid-v4-would-be-here"})

    mock_get.side_effect = [
        MockRequestsResponse(
            {
                "status": "ok",
                "claim": {
                    "createdAt": "2021-08-13T15:18:50.420Z",
                    "updatedAt": "2021-08-13T15:18:58.249Z",
                    "status": FetchAIFaucetApi.FAUCET_STATUS_PENDING,
                },
            }
        ),
        MockRequestsResponse(
            {
                "status": "ok",
                "claim": {
                    "createdAt": "2021-08-13T15:18:50.420Z",
                    "updatedAt": "2021-08-13T15:18:58.249Z",
                    "status": FetchAIFaucetApi.FAUCET_STATUS_PENDING,
                },
            }
        ),
        MockRequestsResponse(
            {
                "status": "ok",
                "claim": {
                    "createdAt": "2021-08-13T15:18:50.420Z",
                    "updatedAt": "2021-08-13T15:18:58.249Z",
                    "status": FetchAIFaucetApi.FAUCET_STATUS_COMPLETED,
                    "txStatus": {
                        "hash": "0x transaction hash would be here",
                        "height": 123456,
                    },
                },
            }
        ),
    ]

    faucet = FetchAIFaucetApi(poll_interval=0)
    faucet.get_wealth(address)

    mock_post.assert_has_calls(
        [
            call(
                url=f"{FetchAIFaucetApi.testnet_faucet_url}/api/v3/claims",
                json={"address": address},
            )
        ]
    )
    mock_get.assert_has_calls(
        [
            call(
                f"{FetchAIFaucetApi.testnet_faucet_url}/api/v3/claims/a-uuid-v4-would-be-here"
            ),
            call(
                f"{FetchAIFaucetApi.testnet_faucet_url}/api/v3/claims/a-uuid-v4-would-be-here"
            ),
            call(
                f"{FetchAIFaucetApi.testnet_faucet_url}/api/v3/claims/a-uuid-v4-would-be-here"
            ),
        ]
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_format_default():
    """Test if default CosmosSDK transaction is correctly formatted."""
    account = FetchAICrypto()
    cc2 = FetchAICrypto()
    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    amount = 10000

    transfer_transaction = cosmos_api.get_transfer_transaction(
        sender_address=account.address,
        destination_address=cc2.address,
        amount=amount,
        tx_fee=1000,
        tx_nonce="something",
        account_number=1,
        sequence=0,
    )

    signed_transaction = account.sign_transaction(transfer_transaction)

    assert "tx" in signed_transaction
    assert "signatures" in signed_transaction["tx"]
    assert len(signed_transaction["tx"]["signatures"]) == 1

    assert "publicKey" in signed_transaction["tx"]["authInfo"]["signerInfos"][0]
    assert "key" in signed_transaction["tx"]["authInfo"]["signerInfos"][0]["publicKey"]
    base64_pbk = signed_transaction["tx"]["authInfo"]["signerInfos"][0]["publicKey"][
        "key"
    ]

    pbk = base64.b64decode(base64_pbk)
    assert pbk == bytes.fromhex(account.public_key)
    assert len(signed_transaction["tx"]["signatures"][0]) == 88
    signature = base64.b64decode(signed_transaction["tx"]["signatures"][0])
    assert len(signature) == 64


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_storage_transaction_cosmwasm():
    """Test the get storage transaction method."""
    cc2 = FetchAICrypto()
    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    contract_interface = {"wasm_byte_code": "1234"}
    deployer_address = cc2.address
    deploy_transaction = cosmos_api.get_deploy_transaction(
        contract_interface,
        deployer_address,
        account_number=1,
        sequence=0,
    )

    assert type(deploy_transaction) == dict and len(deploy_transaction) == 2
    # Check sign_data
    assert "account_number" in deploy_transaction["sign_data"][cc2.address]
    assert "chain_id" in deploy_transaction["sign_data"][cc2.address]

    # Check msg
    assert len(deploy_transaction["tx"]["body"]["messages"]) == 1
    msg = deploy_transaction["tx"]["body"]["messages"][0]
    assert "@type" in msg and msg["@type"] == "/cosmwasm.wasm.v1.MsgStoreCode"

    assert msg["sender"] == deployer_address
    assert msg["wasmByteCode"] == contract_interface["wasm_byte_code"]


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_init_transaction_cosmwasm():
    """Test the get deploy transaction method."""
    cc2 = FetchAICrypto()
    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    init_msg = "init_msg"
    code_id = 1
    deployer_address = cc2.address
    tx_fee = 1
    amount = 10
    gas_limit = 1234
    contract_interface = {}
    init_transaction = cosmos_api.get_deploy_transaction(
        contract_interface,
        deployer_address,
        code_id=code_id,
        init_msg=init_msg,
        amount=amount,
        tx_fee=tx_fee,
        label="",
        account_number=1,
        sequence=0,
        gas=gas_limit,
        denom="abc",
        tx_fee_denom="def",
    )

    assert type(init_transaction) == dict and len(init_transaction) == 2

    # Check sign_data
    assert "account_number" in init_transaction["sign_data"][cc2.address]
    assert "chain_id" in init_transaction["sign_data"][cc2.address]

    # Check tx
    assert init_transaction["tx"]["authInfo"]["fee"]["amount"] == [
        {"denom": "def", "amount": str(tx_fee)}
    ]

    # Check msg
    assert len(init_transaction["tx"]["body"]["messages"]) == 1
    msg = init_transaction["tx"]["body"]["messages"][0]
    assert "@type" in msg and msg["@type"] == "/cosmwasm.wasm.v1.MsgInstantiateContract"
    assert msg["sender"] == deployer_address
    assert msg["codeId"] == str(code_id)
    assert base64.b64decode(msg["msg"]).decode() == f'"{init_msg}"'
    assert msg["funds"] == [{"denom": "abc", "amount": str(amount)}]


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_handle_transaction_cosmwasm():
    """Test the get deploy transaction method."""
    cc2 = FetchAICrypto()

    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    handle_msg = "handle_msg"
    sender_address = cc2.address
    contract_address = "contract_address"
    tx_fee = 1
    amount = 10
    gas_limit = 1234
    handle_transaction = cosmos_api.get_handle_transaction(
        sender_address,
        contract_address,
        handle_msg,
        amount,
        tx_fee,
        gas=gas_limit,
        memo="memo",
        account_number=1,
        sequence=0,
        denom="abc",
        tx_fee_denom="def",
    )

    assert type(handle_transaction) == dict and len(handle_transaction) == 2

    # Check sign_data
    assert "account_number" in handle_transaction["sign_data"][cc2.address]
    assert "chain_id" in handle_transaction["sign_data"][cc2.address]

    # Check tx
    assert handle_transaction["tx"]["authInfo"]["fee"] == {
        "amount": [{"denom": "def", "amount": str(tx_fee)}],
        "gasLimit": str(gas_limit),
    }

    assert "memo" in handle_transaction["tx"]["body"]

    # Check msg
    assert len(handle_transaction["tx"]["body"]["messages"]) == 1
    msg = handle_transaction["tx"]["body"]["messages"][0]
    assert "@type" in msg and msg["@type"] == "/cosmwasm.wasm.v1.MsgExecuteContract"
    assert msg["sender"] == sender_address
    assert msg["contract"] == contract_address
    assert base64.b64decode(msg["msg"]).decode() == f'"{handle_msg}"'
    assert msg["funds"] == [{"denom": "abc", "amount": str(amount)}]


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_try_execute_wasm_query():
    """Test the execute wasm query method."""
    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    client_mock = mock.Mock()

    output_raw_mock = mock.Mock()
    output_raw_mock.data = '{"output": 1}'

    attrs = {"SmartContractState.return_value": output_raw_mock}
    client_mock.configure_mock(**attrs)
    cosmos_api.wasm_client = client_mock
    result = cosmos_api.execute_contract_query(
        contract_address="contract_address", query_msg={}
    )
    assert result == json.loads(output_raw_mock.data)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_send_signed_transaction():
    """Test the send_signed_transaction method"""
    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    tx_signed = cosmos_api.get_transfer_transaction(
        sender_address="addr1",
        destination_address="addr2",
        amount=123,
        tx_fee=1000,
        tx_nonce="something",
        account_number=1,
        sequence=0,
    )

    # Mock version of protobuf Tx response
    mock_return_value = mock.Mock()
    mock_tx_response = mock.Mock()
    mock_tx_response.code = 0
    mock_tx_response.txhash = "digest"
    mock_return_value.tx_response = mock_tx_response

    with mock.patch.object(
        cosmos_api.tx_client, "BroadcastTx", return_value=mock_return_value
    ):
        result = cosmos_api.send_signed_transaction(tx_signed=tx_signed)
    assert result == "digest"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_contract_instance():
    """Test the get contract instance method."""
    cosmos_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)
    assert cosmos_api.get_contract_instance("interface") is None


def test_helper_get_code_id():
    """Test CosmosHelper.is_transaction_settled."""
    assert (
        FetchAIApi.get_code_id(
            {
                "logs": [
                    {
                        "msg_index": 0,
                        "log": "",
                        "events": [
                            {
                                "type": "message",
                                "attributes": [
                                    {"key": "action", "value": "store-code"},
                                    {"key": "module", "value": "wasm"},
                                    {
                                        "key": "signer",
                                        "value": "fetch1pa7q6urt98dfe2rsvfaefj8zhh792sdfuzym2t",
                                    },
                                    {"key": "code_id", "value": "631"},
                                ],
                            }
                        ],
                    }
                ]
            }
        )
        == 631
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_helper_get_contract_address():
    """Test CosmosHelper.is_transaction_settled."""
    assert (
        FetchAIApi.get_contract_address(
            {
                "logs": [
                    {
                        "msg_index": 0,
                        "log": "",
                        "events": [
                            {
                                "type": "message",
                                "attributes": [
                                    {"key": "action", "value": "instantiate"},
                                    {"key": "module", "value": "wasm"},
                                    {
                                        "key": "signer",
                                        "value": "fetch1pa7q6urt98dfe2rsvfaefj8zhh792sdfuzym2t",
                                    },
                                    {"key": "code_id", "value": "631"},
                                    {
                                        "key": "_contract_address",
                                        "value": "fetch1lhd5t8jdjn0n4q27hsah6c0907nxrswcp5l4nw",
                                    },
                                ],
                            }
                        ],
                    }
                ]
            }
        )
        == "fetch1lhd5t8jdjn0n4q27hsah6c0907nxrswcp5l4nw"
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_load_contract_interface():
    """Test the load_contract_interface method."""
    path = Path(ROOT_DIR, "tests", "data", "dummy_contract", "build", "some.wasm")
    result = FetchAIApi.load_contract_interface(path)
    assert "wasm_byte_code" in result


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_construct_init_transaction():
    """Test the construction of a contract instantiate transaction"""
    account = FetchAICrypto()
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    init_transaction = fetchai_api._get_init_transaction(
        deployer_address=account.address,
        denom="atestfet",
        chain_id="cosmoshub-3",
        account_number=1,
        sequence=1,
        amount=0,
        code_id=200,
        init_msg={},
        label="something",
        tx_fee_denom="stake",
    )
    assert (
        isinstance(init_transaction, dict) and len(init_transaction) == 2
    ), "Incorrect transfer_transaction constructed."
    assert (
        init_transaction["tx"]["body"]["messages"][0]["@type"]
        == "/cosmwasm.wasm.v1.MsgInstantiateContract"
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_construct_handle_transaction():
    """Test the construction of a transfer transaction."""
    account = FetchAICrypto()
    account2 = FetchAICrypto()
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    transaction = fetchai_api.get_handle_transaction(
        sender_address=account.address,
        contract_address=account2.address,
        handle_msg={},
        amount=0,
        tx_fee=100,
        denom="atestfet",
        account_number=1,
        sequence=0,
    )
    assert (
        isinstance(transaction, dict) and len(transaction) == 2
    ), "Incorrect transfer_transaction constructed."
    assert (
        transaction["tx"]["body"]["messages"][0]["@type"]
        == "/cosmwasm.wasm.v1.MsgExecuteContract"
    )


def test_load_errors():
    """Test load errors: bad password, no password specified."""
    ec = FetchAICrypto()
    with patch.object(FetchAICrypto, "load", return_value="bad sTring"):
        with pytest.raises(KeyIsIncorrect, match="Try to specify `password`"):
            ec.load_private_key_from_path("any path")

        with pytest.raises(KeyIsIncorrect, match="Wrong password?"):
            ec.load_private_key_from_path("any path", password="some")


def test_decrypt_error():
    """Test bad password error on decrypt."""
    ec = FetchAICrypto()
    ec._pritvate_key = FetchAICrypto.generate_private_key()
    password = "test"
    encrypted_data = ec.encrypt(password=password)
    with patch(
        "aea_ledger_cosmos.cosmos.DataEncrypt.decrypt",
        side_effect=UnicodeDecodeError("expected", b"", 2, 3, ""),
    ):
        with pytest.raises(ValueError, match="bad password?"):
            ec.decrypt(encrypted_data, password + "some")


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_multiple_signatures_transaction():
    """Test generating message with multiple signers"""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    coins = [Coin(denom="DENOM", amount="1234")]

    msg_send = MsgSend(
        from_address=str("from"),
        to_address=str("to"),
        amount=coins,
    )
    send_msg_packed = ProtoAny()
    send_msg_packed.Pack(msg_send, type_url_prefix="/")

    tx = fetchai_api._get_transaction(
        account_numbers=[1, 2],
        from_addresses=["adr1", "adr2"],
        pub_keys=[b"1", b"2"],
        chain_id="chain_id",
        tx_fee=coins,
        gas=1234,
        memo="MEMO",
        sequences=[1, 2],
        msgs=[send_msg_packed, send_msg_packed],
    )
    assert (
        isinstance(tx, dict) and len(tx) == 2
    ), "Incorrect transfer_transaction constructed."
    assert tx["tx"]["body"]["messages"][0]["@type"] == "/cosmos.bank.v1beta1.MsgSend"
    assert tx["tx"]["body"]["messages"][1]["@type"] == "/cosmos.bank.v1beta1.MsgSend"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_multiple_signatures_transaction_missing_pubkeys():
    """Test if generating message with multiple signers without pubkeys fails"""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    coins = [Coin(denom="DENOM", amount="1234")]

    msg_send = MsgSend(
        from_address=str("from"),
        to_address=str("to"),
        amount=coins,
    )
    send_msg_packed = ProtoAny()
    send_msg_packed.Pack(msg_send, type_url_prefix="/")

    with pytest.raises(
        RuntimeError,
        match="Only transaction with one signer can be generated without pubkeys",
    ):
        fetchai_api._get_transaction(
            account_numbers=[1, 2],
            from_addresses=["adr1", "adr2"],
            chain_id="chain_id",
            tx_fee=coins,
            gas=1234,
            memo="MEMO",
            sequences=[1, 2],
            msgs=[send_msg_packed, send_msg_packed],
        )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_multiple_signatures_transaction_wrong_number_of_params():
    """Test if generating message with wrong number of params fails"""
    fetchai_api = FetchAIApi(**FETCHAI_TESTNET_CONFIG)

    coins = [Coin(denom="DENOM", amount="1234")]

    msg_send = MsgSend(
        from_address=str("from"),
        to_address=str("to"),
        amount=coins,
    )
    send_msg_packed = ProtoAny()
    send_msg_packed.Pack(msg_send, type_url_prefix="/")

    with pytest.raises(
        RuntimeError,
        match="Amount of provided from_addresses, sequences and account_numbers is not equal",
    ):
        fetchai_api._get_transaction(
            account_numbers=[1, 2],
            from_addresses=["adr1", "adr2"],
            chain_id="chain_id",
            tx_fee=coins,
            gas=1234,
            memo="MEMO",
            sequences=[1, 2, 3],
            pub_keys=[b"123"],
            msgs=[send_msg_packed],
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Conftest module for Pytest."""
import inspect
import os

from aea_ledger_fetchai import FetchAICrypto


CUR_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))  # type: ignore
ROOT_DIR = os.path.join(CUR_PATH, "..")
MAX_FLAKY_RERUNS = 3
FETCHAI = FetchAICrypto.identifier


FETCHAI_DEFAULT_ADDRESS = "https://rest-dorado.fetch.ai:443"
FETCHAI_DEFAULT_CURRENCY_DENOM = "atestfet"
FETCHAI_DEFAULT_CHAIN_ID = "dorado-1"
FETCHAI_TESTNET_CONFIG = {"address": FETCHAI_DEFAULT_ADDRESS}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Fetchai module wrapping the public and private key cryptography and ledger api."""

from typing import Any, Dict, List, Optional

from aea_ledger_cosmos.cosmos import (
    CosmosCrypto,
    CosmosFaucetApi,
    CosmosHelper,
    MAXIMUM_GAS_AMOUNT,
    _CosmosApi,
)

from aea.common import JSONLike


_ = MAXIMUM_GAS_AMOUNT
_FETCHAI = "fetchai"
_FETCH = "fetch"
TESTNET_NAME = "testnet"
FETCHAI_TESTNET_FAUCET_URL = "https://faucet-dorado.fetch.ai"
DEFAULT_ADDRESS = "https://rest-dorado.fetch.ai:443"
DEFAULT_CURRENCY_DENOM = "atestfet"
DEFAULT_CHAIN_ID = "dorado-1"


class FetchAIHelper(CosmosHelper):
    """Helper class usable as Mixin for FetchAIApi or as standalone class."""

    address_prefix = _FETCH


class FetchAICrypto(CosmosCrypto):  # pylint: disable=W0223
    """Class wrapping the Entity Generation from Fetch.AI ledger."""

    identifier = _FETCHAI
    helper = FetchAIHelper


class FetchAIApi(_CosmosApi, FetchAIHelper):
    """Class to interact with the Fetch ledger APIs."""

    identifier = _FETCHAI

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the Fetch.ai ledger APIs."""
        if "address" not in kwargs:
            kwargs["address"] = DEFAULT_ADDRESS  # pragma: nocover
        if "denom" not in kwargs:
            kwargs["denom"] = DEFAULT_CURRENCY_DENOM
        if "chain_id" not in kwargs:
            kwargs["chain_id"] = DEFAULT_CHAIN_ID
        super().__init__(**kwargs)

    def contract_method_call(
        self,
        contract_instance: Any,
        method_name: str,
        **method_args: Any,
    ) -> Optional[JSONLike]:
        """Call a contract's method

        :param contract_instance: the contract to use
        :param method_name: the contract method to call
        :param method_args: the contract call parameters
        """
        raise NotImplementedError  # pragma: nocover

    def build_transaction(
        self,
        contract_instance: Any,
        method_name: str,
        method_args: Optional[Dict],
        tx_args: Optional[Dict],
        raise_on_try: bool = False,
    ) -> Optional[JSONLike]:
        """Prepare a transaction

        :param contract_instance: the contract to use
        :param method_name: the contract method to call
        :param method_args: the contract parameters
        :param tx_args: the transaction parameters
        :param raise_on_try: whether the method will raise or log on error
        """
        raise NotImplementedError  # pragma: nocover

    def get_transaction_transfer_logs(
        self,
        contract_instance: Any,
        tx_hash: str,
        target_address: Optional[str] = None,
    ) -> Optional[JSONLike]:
        """
        Get all transfer events derived from a transaction.

        :param contract_instance: the contract
        :param tx_hash: the transaction hash
        :param target_address: optional address to filter transfer events to just those that affect it
        """
        raise NotImplementedError  # pragma: nocover

    def send_signed_transactions(
        self,
        signed_transactions: List[JSONLike],
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[List[str]]:
        """
        Simulate and send a bundle of transactions.

        This operation is not supported for fetchai.

        :param signed_transactions: the raw signed transactions to bundle together and send.
        :param raise_on_try: whether the method will raise or log on error.
        :param kwargs: the keyword arguments.
        """
        raise NotImplementedError(  # pragma: nocover
            f"Sending a bundle of transactions is not supported for the {self.identifier} plugin"
        )

    def filter_event(
        self,
        event: Any,
        match_single: Dict[str, Any],
        match_any: Dict[str, Any],
        to_block: int,
        from_block: int,
        batch_size: int,
        max_retries: int,
        reduce_factor: float,
        timeout: int,
    ) -> Optional[JSONLike]:
        """Filter an event using batching to avoid RPC timeouts.

        :param event: the event to filter for.
        :param match_single: the filter parameters with value checking against the event abi. It allows for defining a single match value.
        :param match_any: the filter parameters with value checking against the event abi. It allows for defining multiple match values.
        :param to_block: the block to which to filter.
        :param from_block: the block from which to start filtering.
        :param batch_size: the blocks' batch size of the filtering.
        :param max_retries: the maximum number of retries.
        :param reduce_factor: the percentage by which the batch size is reduced in case of a timeout.
        :param timeout: a timeout in seconds to interrupt the operation in case the RPC request hangs.
        """
        raise NotImplementedError(  # pragma: nocover
            f"Custom events' filtering is not supported for the {self.identifier} plugin"
        )


class FetchAIFaucetApi(CosmosFaucetApi):
    """Fetchai testnet faucet API."""

    identifier = _FETCHAI
    testnet_name = TESTNET_NAME
    testnet_faucet_url = FETCHAI_TESTNET_FAUCET_URL



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the fetchai._cosmos module."""
from unittest.mock import Mock, patch

import pytest
from aea_ledger_fetchai import CosmosHelper, FetchAIApi, FetchAICrypto
from aea_ledger_fetchai.fetchai import MAXIMUM_GAS_AMOUNT
from cosmpy.auth.rest_client import QueryAccountResponse

from tests.conftest import MAX_FLAKY_RERUNS


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_is_tx_settled() -> None:
    """Test is_tx_settled."""
    assert FetchAIApi.is_transaction_settled({"code": None})
    assert FetchAIApi.is_transaction_settled({})
    assert not FetchAIApi.is_transaction_settled({"code": False})


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_is_transaction_valid() -> None:
    """Test is_transaction_valid."""
    assert not CosmosHelper.is_transaction_valid(None, 1, 2, 3, 4)
    tx_good = {
        "tx": {
            "body": {
                "messages": [
                    {"fromAddress": 1, "toAddress": 2, "amount": [{"amount": 4}]}
                ]
            }
        }
    }
    assert CosmosHelper.is_transaction_valid(tx_good, 2, 1, 3, 4)
    assert not CosmosHelper.is_transaction_valid(tx_good, 2, 2, 3, 4)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_transfer_transaction() -> None:
    """Test get_transfer_transaction."""
    fetchai_api = FetchAIApi()
    with patch.object(
        fetchai_api, "_try_get_account_number_and_sequence", return_value=[100, 200]
    ):
        transfer_tx = fetchai_api.get_transfer_transaction(
            "2", "1", 4, 6, 5, gas=MAXIMUM_GAS_AMOUNT + 1
        )
    assert transfer_tx["tx"]["authInfo"]["fee"]["gasLimit"] == f"{MAXIMUM_GAS_AMOUNT}"
    assert CosmosHelper.is_transaction_valid(transfer_tx, "1", "2", 3, 4)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_packed_send_msg() -> None:
    """Test get_packed_send_msg."""
    assert (
        "cosmos.bank.v1beta1.MsgSend"
        in FetchAIApi().get_packed_send_msg(1, 2, 3, b"4").type_url
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_multi_transaction() -> None:
    """Test get_multi_transaction."""
    crypto = FetchAICrypto()
    fetchai_api = FetchAIApi()

    with patch.object(
        fetchai_api, "_try_get_account_number_and_sequence", return_value=[100, 200]
    ):
        result = fetchai_api.get_multi_transaction(
            from_addresses=[crypto.address],
            pub_keys=[bytes.fromhex(crypto.public_key)],
            msgs=[],
            gas=12,
        )
        assert "tx" in result


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_packed_exec_msg() -> None:
    """Test get_packed_exec_msg."""
    crypto = FetchAICrypto()
    fetchai_api = FetchAIApi()

    with patch.object(
        fetchai_api, "_try_get_account_number_and_sequence", return_value=[100, 200]
    ):
        result = fetchai_api.get_packed_exec_msg(
            contract_address=crypto.address,
            sender_address=crypto.address,
            msg={"some": 12},
        )
        assert "MsgExecuteContract" in result.type_url


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_try_get_account_number_and_sequence() -> None:
    """Test _try_get_account_number_and_sequence."""
    fetchai_api = FetchAIApi()
    resp = b"\nV\n /cosmos.auth.v1beta1.BaseAccount\x122\n,fetch1k9dns2fd74644g0q9mfpsmfeqg0h2ym2cm6wdh\x18\xfc\x80\x04"
    msg = QueryAccountResponse()
    msg.ParseFromString(resp)
    with patch.object(fetchai_api.auth_client, "Account", return_value=msg):
        assert fetchai_api._try_get_account_number_and_sequence(
            address="fetch1k9dns2fd74644g0q9mfpsmfeqg0h2ym2cm6wdh"
        ) == (65660, 0)


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_transaction() -> None:
    """Test get_transaction."""
    fetchai_api = FetchAIApi(address="https://rest-fetchhub.fetch.ai:443")
    assert "tx" in fetchai_api.get_transaction(
        "A17E2EEA507D00081826031565F9870771733BBFB62F31FBC356D8AAAC8CBAC8"
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_send_signed_transaction() -> None:
    """Test send_signed_transaction."""
    fetchai_api = FetchAIApi()
    resp_mock = Mock()
    resp_mock.tx_response.code = -1
    with patch.object(fetchai_api.tx_client, "BroadcastTx", return_value=resp_mock):
        assert fetchai_api.send_signed_transaction({"tx": {"body": {}}}) is None

    resp_mock.tx_response.code = 0
    with patch.object(fetchai_api.tx_client, "BroadcastTx", return_value=resp_mock):
        assert fetchai_api.send_signed_transaction({"tx": {"body": {}}}) is not None


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_transaction_receipt() -> None:
    """Test get_transaction_receipt."""
    fetchai_api = FetchAIApi(address="https://rest-fetchhub.fetch.ai:443")
    assert "tx" in fetchai_api.get_transaction_receipt(
        "A17E2EEA507D00081826031565F9870771733BBFB62F31FBC356D8AAAC8CBAC8"
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_handle_transaction() -> None:
    """Test get_handle_transaction."""
    fetchai_api = FetchAIApi()
    with patch.object(
        fetchai_api, "_try_get_account_number_and_sequence", return_value=[100, 200]
    ):
        result = fetchai_api.get_handle_transaction(
            sender_address="1", contract_address="2", handle_msg="", amount=3, tx_fee=4
        )

        assert "tx" in result


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
def test_get_deploy_transaction() -> None:
    """Test get_deploy_transaction."""
    fetchai_api = FetchAIApi()
    with patch.object(
        fetchai_api, "_try_get_account_number_and_sequence", return_value=[100, 200]
    ):
        result = fetchai_api.get_deploy_transaction(
            contract_interface={"wasm_byte_code": "some"},
            deployer_address="Address",
        )

        assert "tx" in result



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains testing utilities."""
import logging
import os
import tempfile
import time
from typing import Any, Dict, Optional


try:
    from docker import DockerClient
    from docker.models.containers import Container
except ImportError:  # pragma: nocover
    DockerClient = Any
    Container = Any


from aea.exceptions import enforce
from aea.helpers import http_requests as requests
from aea.test_tools.docker_image import DockerImage


logger = logging.getLogger(__name__)


class FetchLedgerDockerImage(DockerImage):
    """Wrapper to Fetch ledger Docker image."""

    tmpdirname: tempfile.TemporaryDirectory

    PORTS = {1317: 1317, 26657: 26657}

    def __init__(
        self,
        client: DockerClient,
        addr: str,
        port: int,
        tag: str,
        config: Optional[Dict] = None,
    ):
        """
        Initialize the Fetch ledger Docker image.

        :param client: the Docker client.
        :param addr: the address.
        :param port: the port.
        :param tag: the tag
        :param config: optional configuration to command line.
        """
        super().__init__(client)
        self._addr = addr
        self._port = port
        self._image_tag = tag
        self._config = config or {}

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return self._image_tag

    def _make_entrypoint_file(self, tmpdirname) -> None:
        """Make a temporary entrypoint file to setup and run the test ledger node"""
        run_node_lines = (
            "#!/usr/bin/env bash",
            # variables
            f'export VALIDATOR_KEY_NAME={self._config["genesis_account"]}',
            f'export VALIDATOR_MNEMONIC="{self._config["mnemonic"]}"',
            'export PASSWORD="12345678"',
            f'export CHAIN_ID={self._config["chain_id"]}',
            f'export MONIKER={self._config["moniker"]}',
            f'export DENOM={self._config["denom"]}',
            # Add key
            '( echo "$VALIDATOR_MNEMONIC"; echo "$PASSWORD"; echo "$PASSWORD"; ) |fetchd keys add $VALIDATOR_KEY_NAME --recover',
            # Configure node
            "fetchd init --chain-id=$CHAIN_ID $MONIKER",
            'echo "$PASSWORD" |fetchd add-genesis-account $(fetchd keys show $VALIDATOR_KEY_NAME -a) 100000000000000000000000$DENOM',
            'echo "$PASSWORD" |fetchd gentx $VALIDATOR_KEY_NAME 10000000000000000000000$DENOM --chain-id $CHAIN_ID',
            "fetchd collect-gentxs",
            # Enable rest-api
            'sed -i "s/stake/atestfet/" ~/.fetchd/config/genesis.json',
            'sed -i "s/enable = false/enable = true/" ~/.fetchd/config/app.toml',
            'sed -i "s/swagger = false/swagger = true/" ~/.fetchd/config/app.toml',
            "fetchd start",
        )

        entrypoint_file = os.path.join(tmpdirname, "run-node.sh")
        with open(entrypoint_file, "w") as file:
            file.writelines(line + "\n" for line in run_node_lines)
        os.chmod(entrypoint_file, 300)  # nosec

    def create(self) -> Container:
        """Create the container."""
        self.tmpdirname = tempfile.TemporaryDirectory()
        self._make_entrypoint_file(self.tmpdirname.name)
        mount_path = "/mnt"
        volumes = {self.tmpdirname.name: {"bind": mount_path, "mode": "rw"}}
        entrypoint = os.path.join(mount_path, "run-node.sh")
        container = self._client.containers.run(
            self.tag,
            detach=True,
            network="host",
            volumes=volumes,
            entrypoint=str(entrypoint),
            ports=self.PORTS,
        )
        return container

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""
        for i in range(max_attempts):
            try:
                url = f"{self._addr}:{self._port}/net_info?"
                response = requests.get(url)
                enforce(response.status_code == 200, "")
                return True
            except Exception:  # pylint: disable=broad-except
                logger.info(
                    "Attempt %s failed. Retrying in %s seconds...", i, sleep_rate
                )
                time.sleep(sleep_rate)
        return False  # pragma: nocover
